#!/usr/bin/env Rscript
library(limma)


#' Performs differential expression analysis of multiple groups using limma
#'
#' @param data_frame A matrix-like data object containing log-expression values
#'    for a series of samples.
#' @param experimental_design A matrix-like data object describing the
#'    experimental design of the data_frame, each row in experimental_design
#'    corresponds to a column in data_frame. Must contain a column "Experiment",
#'    which is used for creating the design matrix for the linear model. If
#'    batch correction should be applied, batches must be described in the
#'    "Batch" column.
#' @param comparison_groups A list of experiment pairs for which the results of
#'    the differential expression analysis is returned. Each experiment pair
#'    must be written as one string with a dash between the experiment names,
#'    for example "Experimen1-Experiment2".
#' @param batch Logical, if true batch effects are considered for the
#'    differential expression analysis. Batches must be specified in the
#'    experimental_design in a "Batch" column.
#' @param trend Logical, if true an intensity-dependent trend is fitted to the
#'    prior variance during calculation of the moderated t-statistics, refer to
#'    limma.eBayes for details.
#' @return Returns a named list of dataframe objects, with names corresponding
#'    to the comparison groups. Each dataframe contains the results of the
#'    differential expression analysis for one comparison group, generated by
#'    the topTable() function. Dataframe columns are "logFC", "AveExpr", "t",
#'    "P.Value", "adj.P.Val", and "B".
.multi_group_limma <- function(
    data_frame,
    experimental_design,
    comparison_groups,
    batch,
    trend) {
  sample_experiments <- factor(experimental_design[["Experiment"]])

  if(batch) {
    batch <- factor(experimental_design[["Batch"]])
    design <- model.matrix(~0 + sample_experiments + batch)
  } else {
    design <- model.matrix(~0 + sample_experiments)
  }

  colnames(design) <- sub(
    pattern = ("sample_experiments"),
    replacement = "",
    colnames(design)
  )
  colnames(design) <- sub(
    pattern = ("batch"),
    replacement = "Batch_",
    colnames(design)
  )
  contrast_matrix <- makeContrasts(
    contrasts = comparison_groups,
    levels = design
  )
  
  fit_lm <- lmFit(data_frame, design)
  fit_contrasts <- contrasts.fit(fit_lm, contrast_matrix)
  fit_ebayes <- eBayes(fit_contrasts, trend = trend)
  
  # Use empty vector for adding group comparisons to prevent using append()
  results <- vector("list", length(comparison_groups))
  names(results) <- comparison_groups
  for (group in comparison_groups) {
    results[[group]] <- topTable(
      fit_ebayes,
      number = Inf,
      coef = group,
      adjust = "BH",
      sort.by = "none")
  }
  return(results)
}


#' Performs differential expression analysis between two groups
#'
#' @param data_frame A matrix-like data object containing log-expression values
#'    for a series of samples.
#' @param groups: A list that contains a group name for each column. List entries must
#'    be equal to 'group1' or 'group2'.
#' @param group1: Experimental group 1
#' @param group2: Experimental group 2, used as the coefficient
#' @param trend Logical, if true an intensity-dependent trend is fitted to the
#'    prior variance during calculation of the moderated t-statistics, refer to
#'    limma.eBayes for details.
#' @return Returns a dataframe which contains the results of the differential expression
#'    analysis, generated by the topTable() function. Dataframe columns are "logFC",
#'    "AveExpr", "t", "P.Value", "adj.P.Val", and "B".
.two_group_limma <- function(data_frame, column_groups, group1, group2, trend) {
  design <- model.matrix(~factor(column_groups, levels=c(group1, group2)))
  colnames(design) <- c(group1, group2)

  fit_lm <- lmFit(data_frame, design)
  fit_ebayes <- eBayes(fit_lm, trend = trend)
  limma_results <- topTable(fit_ebayes, number = Inf, coef = group2,
                            adjust = "BH", sort.by = "none")
  limma_results <- cbind(rownames(limma_results), limma_results)
  colnames(limma_results)[1] <- "id"
  return(limma_results)
}
